﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Windows.Forms;
using HeuristicLab.Common;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.Instances.DataAnalysis.Regression.Matlab;
using HeuristicLab.Problems.Instances.DataAnalysis.Regression.Matlab.Api;
using HeuristicLab.Problems.Instances.DataAnalysis.Regression.Matlab.Api.Types;

namespace HeuristicLab.Problems.Instances.DataAnalysis.Views.Regression.Matlab {
  public partial class RegressionMatlabImportDialog : Form {
    private IMatlabConnector _mlConnector;
    public event EventHandler<EventArgs<string>> StatusChangedEvent;
    public event EventHandler<EventArgs<string>> ErrorOccuredEvent;



    /// <summary>
    /// Path to the Matlab file
    /// </summary>
    public string Path {
      get { return ProblemTextBox.Text; }
    }

    /// <summary>
    /// As loading the data can take much time, loaded values will be stored.
    /// So they have not have to be loaded a secound time.
    /// </summary>
    private Dataset _values;

    public RegressionMatlabImportType ImportType {
      get {
        return new RegressionMatlabImportType() {
          Shuffle = false,
          TrainingPercentage = TrainingTestTrackBar.Value,
          TargetVariable = (String)TargetVariableComboBox.SelectedValue,
          Values = _values
        };
      }
    }

    private IList<KeyValuePair<string, MLDatatype>> _selectedVariables;

    public IEnumerable<KeyValuePair<string, MLDatatype>> SelectedVariables {
      get {
        return _selectedVariables;
      }
    }

    public RegressionMatlabImportDialog() {
      InitializeComponent();

      _selectedVariables = new List<KeyValuePair<string, MLDatatype>>();
      _values = new Dataset();

      this.FormClosed += (s, e) => DisposeMatlabConnector();
      OpenFileButton.Click += OpenFileButtonClick;
      VariablesListBox.ItemCheck += (s, e) => BeginInvoke((MethodInvoker)(() => CheckedItemsChanged(s, e)));
      LoadValuesButton.Click += LoadValuesButtonClick;
      TrainingTestTrackBar.ValueChanged += TrainingTestTrackBarValueChanged;
      StatusChangedEvent += (s, e) => BeginInvoke((MethodInvoker)(() => StatusLabel.Text = e.Value));
      ErrorOccuredEvent += (s, e) => BeginInvoke((MethodInvoker)(() => {
        OkButton.Enabled = false;
        ErrorTextBox.Text = e.Value;
        ErrorTextBox.Visible = true;
      }));
    }

    private void CheckedItemsChanged(object sender, ItemCheckEventArgs e) {
      _selectedVariables.Clear();
      foreach (KeyValuePair<string, MLDatatype> item in VariablesListBox.CheckedItems) {
        if (item.Value == MLDatatype.Timeseries) {
          _selectedVariables.Add(item);
        }
      }
      LoadValuesButton.Enabled = VariablesListBox.CheckedIndices.Count > 0;
    }

    #region Events


    private void LoadValuesButtonClick(object sender, EventArgs e) {
      LoadValuesAsync();
    }

    private void OpenFileButtonClick(object sender, EventArgs e) {
      if (openFileDialog.ShowDialog(this) != DialogResult.OK) {
        return;
      }
      GetVariableNamesAsync();
      ErrorTextBox.Visible = false;
    }

    private void TrainingTestTrackBarValueChanged(object sender, System.EventArgs e) {
      TrainingLabel.Text = "Training: " + TrainingTestTrackBar.Value + " %";
      TestLabel.Text = "Test: " + (TrainingTestTrackBar.Maximum - TrainingTestTrackBar.Value) + " %";
    }
    #endregion
    
    /// <summary>
    /// Fires the ErrorOccuredEvent with the given text message.
    /// </summary>
    /// <param name="message"></param>
    private void ErrorOccured(string message) {
      ErrorOccuredEvent?.Invoke(this, new EventArgs<string>(message));
    }

    /// <summary>
    /// Fires the StatusChangedEvent.
    /// </summary>
    /// <param name="text"></param>
    private void UpdateStatus(string text) {
      StatusChangedEvent?.Invoke(this, new EventArgs<string>(text));
    }

    #region Opening and disposing of the Matlab connector
    private void OpenMatlabConnector() {
      if (_mlConnector == null) {
        _mlConnector = MatlabApiFactory.CreateMatlabConnector();
      }
    }

    private async void DisposeMatlabConnector() {
      UpdateStatus("Closing Matlab");
      await Task.Run(() => {
        if (_mlConnector != null) {
          _mlConnector.Dispose();
        }
      });
    }
    #endregion



    /// <summary>
    /// This method reads all available variables form the selected matlab scirpt and puts them into the list view.
    /// </summary>
    private async void GetVariableNamesAsync() {
      UpdateStatus("Opening Matlab");

      var path = openFileDialog.FileName;
      ProblemTextBox.Text = path;
      var variables = await Task.Run(() => {
        OpenMatlabConnector();
        return _mlConnector.GetVariablesFromScript(path);
      });

      UpdateStatus("Loading variable names");
      var variablesListBoxItems = VariablesListBox.Items;
      variablesListBoxItems.Clear();
      foreach (var item in variables) {
        variablesListBoxItems.Add(item, false);
      }
      UpdateStatus("Done");
    }

    /// <summary>
    /// Loads the values of the selected timeseries values and displays them at the preview control.
    /// </summary>
    private async void LoadValuesAsync() {
      try {
        PreviewDatasetMatrix.Content = new Dataset();

        UpdateStatus("Loading values from Matlab");
        var vals = await Task.Run<List<MLTimeseries>>(() => {
          var values = new List<MLTimeseries>();
          foreach (KeyValuePair<string, MLDatatype> item in VariablesListBox.CheckedItems) {
            if (item.Value == MLDatatype.Timeseries) {
              var val = _mlConnector.GetTimeseries(item.Key);
              if (val != null) {
                values.Add(val);
              }
            }
          }
          return values;
        });

        UpdateStatus("Converting values");
        DisplayLoadedValuesAsync(await Task.Run<MLTimeseries>(() => new MLTimeseries(vals)));
        ErrorTextBox.Text = String.Empty;
        ErrorTextBox.Visible = false;
        OkButton.Enabled = true;
      } catch (Exception ex) {
        if (ex is IOException || ex is InvalidOperationException || ex is ArgumentException) {
          ErrorOccured(ex.Message);
        } else {
          throw;
        }
      }
      UpdateStatus("Done");
    }

    /// <summary>
    /// Displays the values of a given timeseries at the preview dataset matrix.
    /// </summary>
    /// <param name="timeseries"></param>
    private async void DisplayLoadedValuesAsync(MLTimeseries timeseries) {
      PreviewDatasetMatrix.Content = null;
      IEnumerable<string> variableNamesWithType = GetVariableNamesWithType(timeseries);
      var values = await Task.Run<double[,]>(() => GetValuesForDataset(timeseries));
      _values = new Dataset(variableNamesWithType, values);
      PreviewDatasetMatrix.Content = _values;
      SetPossibleTargetVariables(timeseries);
    }

    /// <summary>
    /// Sets the entries of the combobox for the target variable.
    /// </summary>
    /// <param name="timeseries"></param>
    protected void SetPossibleTargetVariables(MLTimeseries timeseries) {
      if (timeseries != null) {
        TargetVariableComboBox.DataSource = timeseries.DataHeader;
      }
    }

    /// <summary>
    /// Returns a list with all selected variable names for the preview dataset.
    /// </summary>
    /// <param name="timeseries"></param>
    /// <returns></returns>
    private IEnumerable<string> GetVariableNamesWithType(MLTimeseries timeseries) {
      IList<string> variableNamesWithType = timeseries.DataHeader.ToList();
      variableNamesWithType.Insert(0, "time");
      return variableNamesWithType;
    }

    /// <summary>
    /// Returns a two dimensional double array which contains the values including the timestamps of the given timeseries.
    /// Array[value, variable] 
    /// If the amount of the data is too big, a OutOfMemoryException exception is being thrown. 
    /// This exception is being cought and a message will be shown at the error message box.
    /// </summary>
    /// <param name="timeseries"></param>
    /// <returns></returns>
    private double[,] GetValuesForDataset(MLTimeseries timeseries) {
      try {
        double[,] datasetValues = new double[timeseries.Count, timeseries.DataHeader.Length + 1];

        int i = 0;
        foreach (var item in timeseries.Data) {
          datasetValues[i, 0] = item.Key;
          for (int j = 1; j < datasetValues.GetLength(1); j++) {
            datasetValues[i, j] = item.Value[j - 1];
          }
          i++;
        }
        return datasetValues;
      } catch (OutOfMemoryException ex) {
        ErrorOccured(ex.Message + " Amount of the given data is too big.");
      }
      return new double[0, timeseries.DataHeader.Length + 1];
    }
  }
}

